/********************************************************************************/
/*         (c) Copyright 2018, Real-Time Innovations, All rights reserved.      */
/*                                                                              */
/* RTI grants Licensee a license to use, modify, compile, and create            */ 
/* derivative works of the software solely for use with RTI Connext DDS.        */ 
/* Licensee may redistribute copies of the software provided that all such      */
/* copies are subject to this license. The software is provided "as is", with   */
/* no warranty of any type, including any warranty for fitness for any purpose. */
/* RTI is under no obligation to maintain or support the software.  RTI shall   */
/* not be liable for any incidental or consequential damages arising out of the */
/* use or inability to use the software.                                        */
/********************************************************************************/
#include "GenericSubscriber.h"
#include <iostream>
#include <args.hxx>
#include <dds/dds.hpp>
#include <rti/config/Version.hpp>
#include <rti/domain/find.hpp>

using DomainArgType = args::ValueFlag<int>;
using VerboseArgType = args::Flag;
using ModeArgType = args::Flag;
using TopicNameArgType = args::ValueFlag<std::string>;

int main(int argc, char **argv)
{
    /* create the parser */
    args::ArgumentParser parser("Example DDS program.",
                                "At least one publisher and one subscriber must be started.\n"
                                "Copyright 2018 Real-Time Innovations, Inc.");

    /* add flags to the parser */
    args::HelpFlag       help(parser, "help", "Display this help menu", {'h', "help"});
    VerboseArgType       verbose(parser, "verbose", "output progess messages", {"v", "verbose"});
    DomainArgType        domain(parser, "DOMAIN", "domain to join", {"d", "domain"}, 1);
    ModeArgType          list(parser, "list", "list topics in domain", {"l", "list"});
    TopicNameArgType     topic(parser, "TOPIC", "Subscribe to topic", {"t", "topic"}, "Hello");

    /* complete parsing definition */
    args::CompletionFlag completion(parser, {"complete"});

    try {
        /* parse what is available in argv */
        parser.ParseCLI(argc, argv);
    } catch (args::Completion e) {
        std::cout << e.what();
        return -1;
    } catch (args::Help) {
        std::cout << parser;
        return 0;
    } catch (args::ParseError e) {
        std::cerr << e.what() << std::endl;
        std::cerr << parser;
        return -2;
    }

    // if we get here, all flag objects are valid
    if (verbose.Get()) {
        std::cout << "Running with the following arguments:" << std::endl;
        std::cout << "    Domain ID........ : " << domain.Get() << std::endl;
        std::cout << "RTI Product Version.. : " << rti::config::product_version().to_string() << std::endl;
    }

    try {
        /* 
        * In order to catch the discovery data and thereby gather the 
        * type information, we must create the entity in the disabled
        * state.  This is an attribute of the factory and is set 
        * statically and globally (i.e. unless changed all entities 
        * from this point forward will be created in a disabled state)
        */
        dds::domain::qos::DomainParticipantFactoryQos qos;
        qos << dds::core::policy::EntityFactory::ManuallyEnable();
        dds::domain::DomainParticipant::participant_factory_qos(qos);

        GenericSubscriber subscriber(domain.Get(), verbose.Get());
        if(list.Get()) {
            subscriber.list_topics();
        }  else { 
            subscriber.receive(topic.Get());
        }
    } catch (const dds::core::Error& ex) {
        std::cerr << "Could not create subscriber (General Error): " << ex.what() << std::endl;
    } catch (const dds::core::InvalidArgumentError &ex) {
        std::cerr << "Could not create subscriber (Invalid Argument): " << ex.what() << std::endl;
    } catch (const dds::core::TimeoutError &ex) {
        std::cerr << "Could not create subscriber (Timeout): " << ex.what() << std::endl;
    } catch (const dds::core::UnsupportedError &ex) {
        std::cerr << "Could not create subscriber (Unsupported): " << ex.what() << std::endl;
    } catch (const dds::core::AlreadyClosedError &ex) {
        std::cerr << "Could not create subscriber (Already Closed): " << ex.what() << std::endl;
    } catch (const dds::core::IllegalOperationError &ex) {
        std::cerr << "Could not create subscriber (Illegal Operation): " << ex.what() << std::endl;
    } catch (const dds::core::NotEnabledError &ex) {
        std::cerr << "Could not create subscriber: (Not Enabled) " << ex.what() << std::endl;
    } catch (const dds::core::PreconditionNotMetError &ex) {
        std::cerr << "Could not create subscriber (Precondition Not Met): " << ex.what() << std::endl;
    } catch (const dds::core::ImmutablePolicyError &ex) {
        std::cerr << "Could not create subscriber (Immutable Policy Error): " << ex.what() << std::endl;
    } catch (const dds::core::InconsistentPolicyError &ex) {
        std::cerr << "Could not create subscriber (Inconsistent Policy): " << ex.what() << std::endl;
    } catch (const dds::core::OutOfResourcesError &ex) {
        std::cerr << "Could not create subscriber (Out of Resources): " << ex.what() << std::endl;
    } catch (const dds::core::InvalidDowncastError &ex) {
        std::cerr << "Could not create subscriber (Invalid Downcast): " << ex.what() << std::endl;
    } catch (const dds::core::NullReferenceError &ex) {
        std::cerr << "Could not create subscriber (Null Reference): " << ex.what() << std::endl;
    } catch (const dds::core::InvalidDataError &ex) {
        std::cerr << "Could not create subscriber (Invalid Data): " << ex.what() << std::endl;
    }

    std::cout << "Exiting normally.\n";

    return 0;
}
